# PostScript initialization 

use strict;
use utf8;

our $b_data;
our $o_asize;
our $o_lsize;
our $o_fsize;
our $o_ts;
our $o_lnum;
our $o_paper;
our $o_lines;
our $o_cols;
our $o_wrap_mark;
our $o_nembed;
our $o_margins;
our $o_fset;
our $o_inverse;
our $o_no_ansi;
our $verbose;
our $o_bookish;
our %o_hd;
our $o_wrap;
our $o_lscape;
our $o_title;
our $o_afset;
our $o_lfont;

our $b_ps;

our $c_had_ansi;
our $g_lines;
our $g_cols;

# The problem here is: it's natural to do all font- and margin-related
# calculations in postscript code, BUT we need at least number of lines
# known in perl code, AND ability to specify either font size, nlines of
# ncolumns. On the other hand, frame drawing code should be completely
# postscript.
#
# To solve this, we will do minimal required calculations here -- this
# includes font size / nlines / margins and paper size -- and we will
# assume that margins to be margins of text area alone. Headings. if any,
# will be draws outside of specified margins.
# This solution is not as bad as it might seem, text with and without
# headings will look just the same in this case.
sub init_pages
{
	my ($fsize, $asize, $lsize);

	my ($pw, $ph, $p_margins) = setup_paper();
	my $d_margins = $p_margins if defined $p_margins;
	$d_margins = ($pw > $ph ? $pw : $ph) / 14 unless defined $d_margins;
	my ($ml, $mr, $mt, $mb) = parse_margins(
			$o_margins,
			parse_margins($d_margins, 0, 0, 0, 0)
	);
	($pw, $ph) = ($ph, $pw) if $o_lscape;
	($fsize, $g_lines, $g_cols) = setup_font_lines($pw - $ml - $mr, $ph - $mt - $mb);

	my $ox = $ml;
	my $oy = $ph - $mt - $fsize;

	#warn "got o_asize=$o_asize, as=$as, fsize=$fsize\n" if $verbose;
	$asize = $o_asize if defined $o_asize;
	$asize = ($ph - $mt - $mb)/60 if not defined $asize;
	#warn "set asize=$asize\n" if $verbose;

	if(defined($o_lsize)) {
		$lsize = $o_lsize;
	} else {
		$lsize = 0.8 * $fsize;
		$lsize = $asize if $lsize > $asize;
	};

	prepare_headings();

	print "%!PS-Adobe-2.0\n";
	if($o_lscape) {
		print "%%BoundingBox: 0 0 $ph $pw\n";
	} else {
		print "%%BoundingBox: 0 0 $pw $ph\n";
	}
	print "%%Orientation: ".($o_lscape ? "Landscape" : "Portrait")."\n";
	print "%%Title: $o_title\n";
	print "%%Pages: (atend)\n";
	print "%%Creator: u2ps\n";
	print "%%CreationDate: ".(localtime)."\n";
	print "%%EndComments\n";
	print "\n";

	include_setup(
		'has-ansi' => !$o_no_ansi,
		'has-wrap' => $o_wrap,
		'has-hdrs' => (keys %o_hd ? 1 : 0),
		'has-inrs' => $o_inverse,
		'has-lnum' => $o_lnum,

		'lscape' => $o_lscape,
		inverse => truefalse($o_inverse),
		wrapmark => truefalse($o_wrap_mark),
		bookish => truefalse($o_bookish),

		ts => $o_ts,
		fset => $o_fset,
		hfset => $o_afset,
		lfont => $o_lfont,
		fsize => $fsize,
		asize => $asize,
		lsize => $lsize,

		ml => $ml,
		mr => $mr,
		mt => $mt,
		mb => $mb,
		pw => $pw,
		ph => $ph
	);
}

sub fini_pages
{
	my $pagecount = shift;

	print "\nshowpage\n";
	print "\n";
	print "%%Trailer\n";
	print "%%Pages: $pagecount\n";
	print "%%EOF\n";
}

sub setup_paper
{
	my $paper;
	my @line;
	my ($pw, $ph);
	my $pdsc = "$b_data/paper";
	my @paper;

	die "Internal check failed: undefined \$paper\n" unless defined $o_paper;

	if(open(PAPER, "<", $pdsc)) {
		while(<PAPER>) {
			next unless /\S/;
			next if /^#/;
			my @data = split(/\s+/);
			unless(@data >= 3) {
				warn "$pdsc:$.: bad line\n";
			} else {
				if(lc($data[0]) eq lc($o_paper)) {
					$pw = $data[1];
					$ph = $data[2];
					$paper = $data[0];
					@paper = @data; shift @paper;
					last;
				}
			}
		}
		close(PAPER);
	} else {
		warn "Can't open paper description file $pdsc\n";
	}
	if(!defined($paper) && $o_paper =~ /(\d+)x(\d+)/) {
		@paper = ($1, $2);
	};

	die "Unknown paper type $o_paper\n" unless @paper;
	return @paper;
}

sub setup_font_lines
{
	my $tw = shift;
	my $th = shift;

	my $fsize;
	my $lines;

	if(!defined($o_fsize) && !defined($o_lines) && !defined($o_cols)) {
		if(!$o_lscape) {
			$o_cols = 80;
		} else {
			$o_lines = 40;
		}
	}

	if(defined($o_fsize)) {
		$fsize = $o_fsize;
		warn "-l ignored as -s was set\n" if defined $o_lines;
		$lines = int($th / $fsize);
	} elsif(defined($o_lines)) {
		$lines = int($o_lines);
		$fsize = $th / $lines;
	} elsif(defined($o_cols)) {
		$fsize = $tw / $o_cols * 2;
		$lines = int($th / $fsize);
	} else {
		die "Internal error: none of -f / -s / -c defined\n";
	}

	my $cols = $tw / $fsize * 2;

	#warn sprintf "text = %ix%i cxr = %ix%i (%.1fx%.1f) fs=%f\n",
	#	$tw, $th,
	#	$cols, $lines,
	#	($cols * $fsize / 2),
	#	($lines * $fsize),
	#	$fsize if $verbose;

	return ($fsize, $lines, $cols);
}

sub parse_margins
{
	my $str = shift;
	my @M = @_;

	return @M unless defined $str;

	my @d = split(/:/, $str);
	my @D;

	if(@d == 4) {
		@D = @d;
	} elsif(@d == 2) {
		@D = ($d[0], $d[0], $d[1], $d[1]);
	} elsif(@d == 1) {
		@D = ($d[0], $d[0], $d[0], $d[0]);
	} else {
		die "bad margins specification \"$str\"\n";
	}

	for(my $i = 0; $i < 4; $i++) {
		if($D[$i] =~ /^\+(\d+)/) {
			$M[$i] += $1;
		} elsif($D[$i] =~ /^-(\d+)/) {
			$M[$i] -= $1;
		} elsif($D[$i] =~ /^=(-?\d+)/) {
			$M[$i] = $1;
		} elsif($D[$i] =~ /^\d+/) {
			$M[$i] = $D[$i];
		} else {
			die "bad margins specification ($D[$i])\n";
		};
	}

	return @M;
}

sub include_setup
{
	my %opts = @_;
	my @setup;
	my @stres;

	read_setup(\@setup, \@stres, \%opts);
	head_setup(\@setup) if keys %o_hd;

	unless($o_nembed) {
		print "%%BeginProlog\n";
		foreach (@stres) { print };
		pass_u2ps_defs(@_);
		print "%%EndProlog\n";
	}

	print "%%BeginSetup\n";
	foreach (@setup) { print };
	print "%%EndSetup\n";
}

# Read setup.ps, substituting variables if necessary, into @setup.
# This is necessary because, first, setup.ps may request some resources
# and all resources should be included before setup, and second,
# include_setup_headings may want to include one more proc.
sub read_setup
{
	my $setup = shift;
	my $stres = shift;
	my $var = shift;
	my $text = 0;
	my $sep = 0;

	my $sfile = "$b_ps/setup.ps";

	open(SETUP, '<', $sfile) || die "Can't open $sfile: $!\n";
	while(<SETUP>) {
		chomp;

		if(/([^%]*[^%\s])\s*%(\S+)\?/) { if( $var->{$2}) { $_ = $1 } else { next } }
		if(/([^%]*[^%\s])\s*%!(\S+)/ ) { if(!$var->{$2}) { $_ = $1 } else { next } }
		s/\$([A-Za-z0-9]+)/defined($var->{$1}) ? $var->{$1} : "null"/ge;
		$_ .= "\n";

		# with -E, we won't run u2ps_pp at all
		if(/^%%IncludeBatch/) {
			push(@$stres, $_) unless $o_nembed;
		} elsif(/^%%NeedsResource/) {
			push(@$stres, $_) unless $o_nembed;
		} elsif(!/\S/) {
			$sep = 1;
		} else {
			push(@$setup, "\n") if $text && $sep;
			push(@$setup, $_);
			$text = 1;
			$sep = 0;
		}
	};
	close(SETUP);
}

sub pass_u2ps_defs
{
	my %opts = @_;
	my $defs = "$b_ps/ProcSet/u2ps.defs";
	my $skip = 0;
	my $text = 0;
	open(DEFS, "<", $defs) || die "Can't open $defs: $!\m";
	print "%%BeginResource: procset u2ps.defs\n";
	while(<DEFS>) {
		if(/^%==\s*(\S+)/) {
			$skip = !$opts{"has-$1"};
			next;
		} elsif(/^%==/) {
			$skip = 0;
			next;
		};
		next if $skip;
		next if /^%/;
		next unless /\S/ || $text;
		$text = 1;
		print;
	}
	print "%%EndResource\n";
	close(DEFS);
}

1;
